@@ -74,6 +74,7 @@
 #include <sys/console.h>
 #include <sys/kobj.h>
 #include <sys/time.h>
+#include <sys/zfs_ioctl.h>
 
 int zevent_len_max = 0;
 int zevent_cols = 80;
@@ -413,7 +414,7 @@ fm_nvprint(nvlist_t *nvl)
 }
 
 static zevent_t *
-fm_event_alloc(void)
+zfs_zevent_alloc(void)
 {
 	zevent_t *ev;
 
@@ -421,49 +422,49 @@ fm_event_alloc(void)
 	if (ev == NULL)
 		return NULL;
 
-	list_create(&ev->ev_zpd_list, sizeof(zfs_private_data_t),
-		    offsetof(zfs_private_data_t, zpd_node));
+	list_create(&ev->ev_ze_list, sizeof(zfs_zevent_t),
+		    offsetof(zfs_zevent_t, ze_node));
 	list_link_init(&ev->ev_node);
 
 	return ev;
 }
 
 static void
-fm_event_free(zevent_t *ev)
+zfs_zevent_free(zevent_t *ev)
 {
 	/* Run provided cleanup callback */
 	ev->ev_cb(ev->ev_nvl, ev->ev_detector);
 
-	list_destroy(&ev->ev_zpd_list);
+	list_destroy(&ev->ev_ze_list);
 	kmem_free(ev, sizeof(zevent_t));
 }
 
 static void
-fm_zevent_drain(zevent_t *ev)
+zfs_zevent_drain(zevent_t *ev)
 {
-	zfs_private_data_t *zpd;
+	zfs_zevent_t *ze;
 
 	ASSERT(MUTEX_HELD(&zevent_lock));
 	list_remove(&zevent_list, ev);
 
 	/* Remove references to this event in all private file data */
-	while ((zpd = list_head(&ev->ev_zpd_list)) != NULL) {
-		list_remove(&ev->ev_zpd_list, zpd);
-		zpd->zpd_zevent = NULL;
-		zpd->zpd_dropped++;
+	while ((ze = list_head(&ev->ev_ze_list)) != NULL) {
+		list_remove(&ev->ev_ze_list, ze);
+		ze->ze_zevent = NULL;
+		ze->ze_dropped++;
 	}
 
-	fm_event_free(ev);
+	zfs_zevent_free(ev);
 }
 
 void
-fm_zevent_drain_all(int *count)
+zfs_zevent_drain_all(int *count)
 {
 	zevent_t *ev;
 
 	mutex_enter(&zevent_lock);
 	while ((ev = list_head(&zevent_list)) != NULL)
-		fm_zevent_drain(ev);
+		zfs_zevent_drain(ev);
 
 	*count = zevent_len_cur;
 	zevent_len_cur = 0;
@@ -478,12 +479,12 @@ fm_zevent_drain_all(int *count)
  * this reference set to NULL.
  */
 static void
-fm_zevent_insert(zevent_t *ev)
+zfs_zevent_insert(zevent_t *ev)
 {
 	mutex_enter(&zevent_lock);
 	list_insert_head(&zevent_list, ev);
 	if (zevent_len_cur >= zevent_len_max)
-		fm_zevent_drain(list_tail(&zevent_list));
+		zfs_zevent_drain(list_tail(&zevent_list));
 	else
 		zevent_len_cur++;
 
@@ -494,7 +495,7 @@ fm_zevent_insert(zevent_t *ev)
  * Post a zevent
  */
 void
-fm_zevent_post(nvlist_t *nvl, nvlist_t *detector, zevent_cb_t *cb)
+zfs_zevent_post(nvlist_t *nvl, nvlist_t *detector, zevent_cb_t *cb)
 {
 	int64_t tv_array[2];
 	timestruc_t tv;
@@ -518,7 +519,7 @@ fm_zevent_post(nvlist_t *nvl, nvlist_t *detector, zevent_cb_t *cb)
 	if (zevent_console)
 		fm_nvprint(nvl);
 
-	ev = fm_event_alloc();
+	ev = zfs_zevent_alloc();
 	if (ev == NULL) {
 		atomic_add_64(&erpt_kstat_data.erpt_dropped.value.ui64, 1);
 		return;
@@ -527,22 +528,56 @@ fm_zevent_post(nvlist_t *nvl, nvlist_t *detector, zevent_cb_t *cb)
         ev->ev_nvl = nvl;
 	ev->ev_detector = detector;
 	ev->ev_cb = cb;
-	fm_zevent_insert(ev);
+	zfs_zevent_insert(ev);
 	cv_broadcast(&zevent_cv);
 }
 
+static int
+zfs_zevent_minor_to_state(minor_t minor, zfs_zevent_t **ze)
+{
+	*ze = zfsdev_get_state(minor, ZST_ZEVENT);
+	if (*ze == NULL)
+		return (EBADF);
+
+	return (0);
+}
+
+int
+zfs_zevent_fd_hold(int fd, minor_t *minorp, zfs_zevent_t **ze)
+{
+	file_t *fp;
+	int error;
+
+        fp = getf(fd);
+        if (fp == NULL)
+                return (EBADF);
+
+        *minorp = zfsdev_getminor(fp->f_file);
+        error = zfs_zevent_minor_to_state(*minorp, ze);
+
+	if (error)
+		zfs_zevent_fd_rele(fd);
+
+	return (error);
+}
+
+void
+zfs_zevent_fd_rele(int fd)
+{
+	releasef(fd);
+}
+
 /*
- * Get the next zevent in the stream.  To avoid making an extra copy of the
- * nvlist we must call put_nvlist() here safely under the zevent_lock.
+ * Get the next zevent in the stream and place a copy in 'event'.
  */
 int
-fm_zevent_next(zfs_private_data_t *zpd, zfs_cmd_t *zc)
+zfs_zevent_next(zfs_zevent_t *ze, nvlist_t **event, uint64_t *dropped)
 {
 	zevent_t *ev;
 	int error;
 
 	mutex_enter(&zevent_lock);
-	if (zpd->zpd_zevent == NULL) {
+	if (ze->ze_zevent == NULL) {
 		/* New stream start at the beginning/tail */
 		ev = list_tail(&zevent_list);
 		if (ev == NULL) {
@@ -552,20 +587,20 @@ fm_zevent_next(zfs_private_data_t *zpd, zfs_cmd_t *zc)
 	} else {
 		/* Existing stream continue with the next element and remove
 		 * ourselves from the wait queue for the previous element */
-		ev = list_prev(&zevent_list, zpd->zpd_zevent);
+		ev = list_prev(&zevent_list, ze->ze_zevent);
 		if (ev == NULL) {
 			error = ENOENT;
 			goto out;
 		}
 
-		list_remove(&zpd->zpd_zevent->ev_zpd_list, zpd);
+		list_remove(&ze->ze_zevent->ev_ze_list, ze);
 	}
 
-	zpd->zpd_zevent = ev;
-	list_insert_head(&ev->ev_zpd_list, zpd);
-	error = put_nvlist(zc, ev->ev_nvl);
-	zc->zc_cookie = zpd->zpd_dropped;
-	zpd->zpd_dropped = 0;
+	ze->ze_zevent = ev;
+	list_insert_head(&ev->ev_ze_list, ze);
+	nvlist_dup(ev->ev_nvl, event, KM_SLEEP);
+	*dropped = ze->ze_dropped;
+	ze->ze_dropped = 0;
 out:
 	mutex_exit(&zevent_lock);
 
@@ -573,7 +608,7 @@ out:
 }
 
 int
-fm_zevent_wait(zfs_private_data_t *zpd)
+zfs_zevent_wait(zfs_zevent_t *ze)
 {
 	int error = 0;
 
@@ -597,23 +632,23 @@ out:
 }
 
 void
-fm_zevent_init(zfs_private_data_t *zpd)
+zfs_zevent_init(zfs_zevent_t **zep)
 {
-	list_link_init(&zpd->zpd_node);
-	zpd->zpd_zevent = NULL;
-	zpd->zpd_dropped = 0;
+	zfs_zevent_t *ze;
+
+	ze = *zep = kmem_zalloc(sizeof (zfs_zevent_t), KM_SLEEP);
+	list_link_init(&ze->ze_node);
 }
 
 void
-fm_zevent_fini(zfs_private_data_t *zpd)
+zfs_zevent_destroy(zfs_zevent_t *ze)
 {
 	mutex_enter(&zevent_lock);
-	if (zpd->zpd_zevent)
-		list_remove(&zpd->zpd_zevent->ev_zpd_list, zpd);
-
-	zpd->zpd_zevent = NULL;
-	zpd->zpd_dropped = 0;
+	if (ze->ze_zevent)
+		list_remove(&ze->ze_zevent->ev_ze_list, ze);
 	mutex_exit(&zevent_lock);
+
+	kmem_free(ze, sizeof (zfs_zevent_t));
 }
 #endif /* _KERNEL */
 
@@ -705,8 +740,8 @@ fm_nvlist_create(nv_alloc_t *nva)
 
 	if (nvlist_xalloc(&nvl, NV_UNIQUE_NAME, nvhdl) != 0) {
 		if (hdl_alloced) {
-			kmem_free(nvhdl, sizeof (nv_alloc_t));
 			nv_alloc_fini(nvhdl);
+			kmem_free(nvhdl, sizeof (nv_alloc_t));
 		}
 		return (NULL);
 	}
@@ -1465,7 +1500,7 @@ fm_fini(void)
 {
 	int count;
 
-	fm_zevent_drain_all(&count);
+	zfs_zevent_drain_all(&count);
 	cv_broadcast(&zevent_cv);
 
 	mutex_enter(&zevent_lock);
