@@ -687,7 +687,7 @@ zfs_ereport_post(const char *subclass, spa_t *spa, vdev_t *vd, zio_t *zio,
 		return;
 
 	/* Cleanup is handled by the callback function */
-	fm_zevent_post(ereport, detector, zfs_zevent_post_cb);
+	zfs_zevent_post(ereport, detector, zfs_zevent_post_cb);
 #endif
 }
 
@@ -739,7 +739,7 @@ zfs_ereport_finish_checksum(zio_cksum_report_t *report,
 	    good_data, bad_data, report->zcr_length, drop_if_identical);
 
 	if (info != NULL)
-		fm_zevent_post(report->zcr_ereport,
+		zfs_zevent_post(report->zcr_ereport,
 		    report->zcr_detector, zfs_zevent_post_cb);
 
 	report->zcr_ereport = report->zcr_detector = NULL;
@@ -771,7 +771,7 @@ void
 zfs_ereport_send_interim_checksum(zio_cksum_report_t *report)
 {
 #ifdef _KERNEL
-	fm_zevent_post(report->zcr_ereport, report->zcr_detector, NULL);
+	zfs_zevent_post(report->zcr_ereport, report->zcr_detector, NULL);
 #endif
 }
 
@@ -795,7 +795,7 @@ zfs_ereport_post_checksum(spa_t *spa, vdev_t *vd,
 	    B_FALSE);
 
 	if (info != NULL) {
-		fm_zevent_post(ereport, detector, zfs_zevent_post_cb);
+		zfs_zevent_post(ereport, detector, zfs_zevent_post_cb);
 		kmem_free(info, sizeof (*info));
 	}
 #endif
@@ -827,7 +827,7 @@ zfs_post_common(spa_t *spa, vdev_t *vd, const char *name)
 		    FM_EREPORT_PAYLOAD_ZFS_VDEV_STATE, vd->vdev_state) == 0);
 	}
 
-	fm_zevent_post(resource, NULL, zfs_zevent_post_cb);
+	zfs_zevent_post(resource, NULL, zfs_zevent_post_cb);
 #endif
 }
 
